@@ -1,18 +1,22 @@
-/* matchbox - a lightweight window manager
-
-   Copyright 2002 Matthew Allum
-
-   This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
-
-   This program is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
-*/
-
+/* 
+ *  Matchbox Window Manager - A lightweight window manager not for the
+ *                            desktop.
+ *
+ *  Authored By Matthew Allum <mallum@o-hand.com>
+ *
+ *  Copyright (c) 2002, 2004 OpenedHand Ltd - http://o-hand.com
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2, or (at your option)
+ *  any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ */
 
 #include "dialog_client.h"
 
@@ -31,12 +35,6 @@
 /* Selected restraint */
 #define DIALOG_DRAG_MODE         DIALOG_DRAG_FREE
 
-/* Border size for DIALOG_DRAG_RESTRAIN */
-#define DIALOG_DRAG_RESTRAIN_BDR 16
-
-/* Hide the dialog ( just show border ) when dragging  */
-#define DIALOG_WANT_HIDDEN_DRAG  0
-
 
 /********************************************/
 
@@ -54,12 +52,11 @@ dialog_client_new(Wm *w, Window win, Client *trans)
 
    if (!c) return NULL;
 
-   c->type = dialog;
+   c->type = MBCLIENT_TYPE_DIALOG;
    
    c->reparent     = &dialog_client_reparent;
    c->move_resize  = &dialog_client_move_resize;
-   c->hide         = &dialog_client_hide;
-   
+   c->iconize      = &dialog_client_iconize;   
    c->configure    = &dialog_client_configure;
    c->button_press = &dialog_client_button_press;
    c->redraw       = &dialog_client_redraw;
@@ -67,19 +64,10 @@ dialog_client_new(Wm *w, Window win, Client *trans)
    c->destroy      = &dialog_client_destroy;
    c->get_coverage = &dialog_client_get_coverage;
 
-   if (trans)
-     {
-       /* 
-        * Dont let dialogs be transient for other dialogs.
-        */
-       if (trans->type != mainwin)
-	 c->trans = trans->trans;
-       else
-	 c->trans = trans;	 
-     }
-
    dialog_client_check_for_state_hints(c);
 
+   c->trans = trans;
+
    return c;
 }
 
@@ -93,8 +81,8 @@ dialog_client_get_offsets(Client *c, int *e, int *s, int *w)
       return;
     }
 
-#ifdef USE_MSG_WIN
-  if (c->flags & CLIENT_IS_MESSAGE_DIALOG)
+  if (c->flags & CLIENT_HAS_URGENCY_FLAG
+      && theme_has_message_decor(c->wm->mbtheme))
     {
       *s = theme_frame_defined_height_get(c->wm->mbtheme, 
 				       FRAME_MSG_SOUTH);
@@ -104,7 +92,6 @@ dialog_client_get_offsets(Client *c, int *e, int *s, int *w)
 					 FRAME_MSG_WEST );
       return;
     }
-#endif 
 
    *s = theme_frame_defined_height_get(c->wm->mbtheme, 
 				       FRAME_DIALOG_SOUTH);
@@ -119,20 +106,22 @@ dialog_client_check_for_state_hints(Client *c)
 {
   if (ewmh_state_check(c, c->wm->atoms[WINDOW_STATE_MODAL]))
     {
-      Client *damaged;
-
+      Client *damaged_client;
       dbg("%s() got modal hint, setting flag\n", __func__);
 
       c->flags ^= CLIENT_IS_MODAL_FLAG;
 
-      /* Call comp_engine_client_show to add damage to main window 
-       * so it gets fully lowlighted ok. 
+      /* Call comp_engine_client_show to add damage to main window
+       * so it gets fully lowlighted ok.
        */
-      if ((damaged = wm_get_visible_main_client(c->wm)) != NULL)
-	{
-	  comp_engine_client_show(c->wm, damaged);
-	}
+      if ((damaged_client = wm_get_visible_main_client(c->wm)) != NULL)
+        {
+          comp_engine_client_show(c->wm, damaged_client);
+        }
     }
+
+  if (ewmh_state_check(c, c->wm->atoms[WINDOW_STATE_ABOVE]))
+    c->flags |= CLIENT_HAS_ABOVE_STATE;
 }
 
 void
@@ -167,17 +156,10 @@ dialog_client_move_resize(Client *c)
 #ifndef USE_COMPOSITE
    if (c->wm->config->dialog_shade && (c->flags & CLIENT_IS_MODAL_FLAG))
      {
-       XMoveResizeWindow(c->wm->dpy, 
-			 c->title_frame, 
-			 c->x - offset_west, 
-			 c->y - frm_size, 
-			 c->width + offset_east + offset_west,
-			 c->height + frm_size + offset_south );
        XMoveWindow(c->wm->dpy, 
 		   c->window, 
 		   c->x, 
 		   c->y);
-
      }
    else
 #endif
@@ -189,142 +171,166 @@ dialog_client_move_resize(Client *c)
 			 c->width + offset_west + offset_east,
 			 c->height + frm_size + offset_south
 			 );
-       XResizeWindow(c->wm->dpy, 
-			 c->title_frame, 
-			 c->width + offset_east + offset_west,
-			 c->height + frm_size + offset_south );
      }
-}
 
-void
-dialog_client_hide(Client *c)
-{
-  Client *damaged;
-  dbg("%s() called for %s\n", __func__, c->name);
-  XLowerWindow(c->wm->dpy, c->frame);
-
-  if (c->flags & CLIENT_IS_MODAL_FLAG 
-      && (damaged = wm_get_visible_main_client(c->wm)) != NULL)
-    {
-      comp_engine_client_show(c->wm, damaged);
-    }
-
-  comp_engine_client_hide(c->wm, c);
+   if (!(c->flags & CLIENT_TITLE_HIDDEN_FLAG))
+     client_decor_frames_move_resize(c, offset_west, offset_east, 
+				     frm_size, offset_south);
 }
 
 int
 dialog_client_title_height(Client *c)
 {
-   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
-      return 0;
+  Wm *w = c->wm;
 
-#ifdef USE_MSG_WIN
-  if (c->flags & CLIENT_IS_MESSAGE_DIALOG)
+  if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
+    return 0;
+  
+  if (c->flags & CLIENT_HAS_URGENCY_FLAG 
+      && theme_has_message_decor(w->mbtheme))
     {
       return theme_frame_defined_height_get(c->wm->mbtheme, FRAME_MSG);
     }
-#endif
-
+  
   if (c->flags & CLIENT_BORDERS_ONLY_FLAG
       && theme_has_frame_type_defined(c->wm->mbtheme, FRAME_DIALOG_NORTH))    
     return theme_frame_defined_height_get(c->wm->mbtheme, FRAME_DIALOG_NORTH);
-
-   return theme_frame_defined_height_get(c->wm->mbtheme, FRAME_DIALOG);
+  
+  return theme_frame_defined_height_get(c->wm->mbtheme, FRAME_DIALOG);
 }
 
+
 void
 dialog_client_show(Client *c)
 {
+  Wm *w = c->wm;
+  Client *highest_client = NULL;
+  MBList *transient_list = NULL, *list_item = NULL;
+
   dbg("%s() called for %s\n", __func__, c->name);
 
-  /* XXX paint before map so dialog gets shape 
-         can probably be more optimised / experimental at the moment. 
-  */
-  if (client_get_state(c) != NormalState)
-    dialog_client_redraw(c, False);
-
-  client_set_state(c, NormalState); 
-  XMapSubwindows(c->wm->dpy, c->frame);
-  XMapRaised(c->wm->dpy, c->frame);
-
-  if (client_want_focus(c) && (!(c->flags & CLIENT_IS_MESSAGE_DIALOG)))
+  if (!c->mapped)
     {
-      XSetInputFocus(c->wm->dpy, c->window,
-		     RevertToPointerRoot, CurrentTime);
-      c->wm->focused_client = c;
+      XMapSubwindows(w->dpy, c->frame);
+      XMapWindow(w->dpy, c->frame);
     }
 
-  comp_engine_client_show(c->wm, c);
+  /* 
+   *  We just need to get the order right in respect 
+   *  to other dialogs transient to the same app or 
+   *  transient to root.
+   *  This order *should* be kept the same by other 
+   *  stack operations ( eg wm_activate client ).
+   */
 
-#ifdef USE_MSG_WIN 	
-  if (c->flags & CLIENT_IS_MESSAGE_DIALOG_LO)
-    {      Client *t = NULL;
+  if (c->trans)
+    {
+      /* Were transient for something 
+       * - so recursives find the highest transient for this app
+       * - raise ourselves above 
+       */
+      Client *lowest_trans = c->trans;
+      int urgent_flag = 0; /* (c->flags & CLIENT_HAS_URGENCY_FLAG) ?
+			      CLIENT_HAS_URGENCY_FLAG : 0; */
 
-      /* Make sure any higher priority dialogs are mapped above */
-      START_CLIENT_LOOP(c->wm, t)
+      while (lowest_trans->trans != NULL) 
+	lowest_trans = lowest_trans->trans;
+
+      highest_client = client_get_highest_transient(lowest_trans, urgent_flag);
+
+      if (c->mapped && highest_client == c)
 	{
-	  if (t->flags & CLIENT_IS_MESSAGE_DIALOG_HI)
+	  /* if were already at the top, logic below will actually
+	   *  move us below the transient.
+	   */
+	  dbg("%s() %s already highest and mapped .. leaving\n", 
+	      __func__, c->name);
+	}
+      else
+	{
+	  if (highest_client == NULL || highest_client == c)
 	    {
-	      t->show(t);
-	      /* We need to call show to get the composite stacking right */
-	      comp_engine_client_show(c->wm, t);
+	      dbg("%s() raising %s above c->trans: %s\n", __func__, 
+		  c->name, c->trans->name);
+	      stack_move_above_client(c, c->trans);
+	    }
+	  else
+	    {
+	      dbg("%s() raising %s above highest_client: %s\n", 
+		  __func__, c->name, highest_client->name);
+	      stack_move_above_client(c, highest_client);
 	    }
 	}
-      END_CLIENT_LOOP(c->wm, t);
+    }
+  else
+    stack_move_top(c);
+
+  /* Now move any transients for us above us */
+
+  client_get_transient_list(w, &transient_list, c);
+ 
+  highest_client = c;
+
+  list_enumerate(transient_list, list_item)
+    {
+      stack_move_above_client((Client *)list_item->data, highest_client);
+      highest_client = (Client *)list_item->data;
     }
 
-  if (c->wm->msg_win_queue_head)
+  list_destroy(&transient_list);
+
+
+  /* Insurance below */
+
+  if (wm_get_visible_main_client(w))
     {
-      Client *msg_client = NULL;
-      if ((msg_client = wm_find_client(c->wm, 
-				       c->wm->msg_win_queue_head->win, 
-				       WINDOW)) != NULL)
-	if (msg_client != c) 
-	  {
-	    msg_client->show(msg_client);
-	    /* We need to call show to get the composite stacking right */
-	    comp_engine_client_show(c->wm, msg_client);
-	  }
+      stack_move_transients_to_top(w, wm_get_visible_main_client(w), 
+				   CLIENT_HAS_ABOVE_STATE);
     }
-#endif
+
+  stack_move_transients_to_top(w, NULL, CLIENT_HAS_ABOVE_STATE);
+
+  stack_move_transients_to_top(w, NULL, CLIENT_HAS_URGENCY_FLAG);
 
   c->mapped = True;
+
 }
 
 void
 dialog_client_reparent(Client *c)
 {
-   XSetWindowAttributes attr;
-
-   int offset_north = dialog_client_title_height(c);
-   int offset_south = 0, offset_west = 0, offset_east = 0;
-
-   dialog_client_get_offsets(c, &offset_east, &offset_south, &offset_west);
-
-   attr.override_redirect = True; 
-   attr.background_pixel  = 0; /* BlackPixel(c->wm->dpy, c->wm->screen); */
-   attr.border_pixel      = 0;
-   attr.event_mask        = ChildMask|ButtonPressMask|ExposureMask;
-
-   attr.colormap          = c->cmap; 
-
-   dbg("%s() want lowlight : wm:%i , client:%i\n", __func__,
-       c->wm->config->dialog_shade, (c->flags & CLIENT_IS_MODAL_FLAG));
+  Wm *w = c->wm;
+  XSetWindowAttributes attr;
+  
+  int offset_north = dialog_client_title_height(c);
+  int offset_south = 0, offset_west = 0, offset_east = 0;
+  
+  dialog_client_get_offsets(c, &offset_east, &offset_south, &offset_west);
+  
+  attr.override_redirect = True; 
+  attr.background_pixel  = w->grey_col.pixel;
+  attr.border_pixel      = 0;
+  attr.event_mask        = ChildMask|ButtonPressMask|ExposureMask;
+  
+  attr.colormap          = c->cmap; 
+  
+  dbg("%s() want lowlight : wm:%i , client:%i\n", __func__,
+      c->wm->config->dialog_shade, (c->flags & CLIENT_IS_MODAL_FLAG));
 #ifndef USE_COMPOSITE
-   if (c->wm->config->dialog_shade && (c->flags & CLIENT_IS_MODAL_FLAG))
-     {
-       dbg("%s() LOWLIGHTING\n", __func__);
-       wm_lowlight(c->wm, c);
-     }
-   else
+  if (c->wm->config->dialog_shade && (c->flags & CLIENT_IS_MODAL_FLAG))
+    {
+      dbg("%s() LOWLIGHTING\n", __func__);
+      wm_lowlight(w, c);
+    }
+  else
 #endif
      {
        if (c->flags & CLIENT_TITLE_HIDDEN_FLAG) 
 	 {
 	   c->frame = c->window;
 	 }
-       else c->frame = XCreateWindow(c->wm->dpy, 
-				     c->wm->root, 
+       else c->frame = XCreateWindow(w->dpy, 
+				     w->root, 
 				     0, 0,
 				     c->width + offset_east + offset_west, 
 				     c->height + offset_north + offset_south, 
@@ -343,36 +349,26 @@ dialog_client_reparent(Client *c)
 				     &attr);
      }
 
-   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
+   if (!(c->flags & CLIENT_TITLE_HIDDEN_FLAG))
      {
-       c->title_frame = c->window;
-     }
-   else
-     {
-       c->title_frame =
-	 XCreateWindow(c->wm->dpy, 
-		       c->frame, 
-		       0, 0, 
-		       c->width + offset_east + offset_west, 
-		       c->height + offset_north + offset_south, 
-		       0,
-		       CopyFromParent, /* depth */
-		       CopyFromParent, 
-		       CopyFromParent, /* visual */
-		       CWOverrideRedirect|CWBackPixel|CWEventMask|CWBorderPixel|CWColormap, 
-		       &attr);
+
+       client_decor_frames_init(c, 
+				offset_west, offset_east, 
+				offset_north, offset_south);
 
      }
+
+   XClearWindow(w->dpy, c->frame);
    
-   XSetWindowBorderWidth(c->wm->dpy, c->window, 0);
+   XSetWindowBorderWidth(w->dpy, c->window, 0);
 
-   XAddToSaveSet(c->wm->dpy, c->window);
+   XAddToSaveSet(w->dpy, c->window);
 
-   XSelectInput(c->wm->dpy, c->window,
-		ButtonPressMask|ColormapChangeMask|PropertyChangeMask);
+   XSelectInput(w->dpy, c->window,
+		ColormapChangeMask|PropertyChangeMask);
 
    if (c->frame != c->window)
-     XReparentWindow(c->wm->dpy, c->window, c->frame, 
+     XReparentWindow(w->dpy, c->window, c->frame, 
 		     offset_west, offset_north);
 }
 
@@ -393,7 +389,7 @@ dialog_get_available_area(Client *c,
   Wm *w = c->wm;
 
   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG 
-      || c->flags & CLIENT_IS_MESSAGE_DIALOG)
+      || c->flags & CLIENT_HAS_URGENCY_FLAG)
     {
       /* Decorationless dialogs can position themselves anywhere */
       *y = 0; *x = 0; *height = w->dpy_height; *width = w->dpy_width;
@@ -403,13 +399,12 @@ dialog_get_available_area(Client *c,
       Client *p = NULL;
       Bool    have_toolbar = False;
 
-      START_CLIENT_LOOP(w, p)
+     stack_enumerate(w, p)
       {
-	if (p->type == toolbar && p->mapped 
+	if (p->type == MBCLIENT_TYPE_TOOLBAR && p->mapped 
 	    && !(p->flags & CLIENT_IS_MINIMIZED))
 	  { have_toolbar = True; break; }
       }
-      END_CLIENT_LOOP(w, p);
 
       *y      = wm_get_offsets_size(w, NORTH, NULL, True);
 
@@ -426,7 +421,7 @@ dialog_get_available_area(Client *c,
 }
 
 /* 
- * dialog_check_gemoetry()
+ * dialog_constrain_gemoetry()
  *
  * called mainly by wm_restack to suggest better positions for dialogs
  * in relation to panels and toolbar/input wins. 
@@ -438,11 +433,11 @@ dialog_get_available_area(Client *c,
  *  to fit. 
  */
 Bool
-dialog_check_geometry(Client *c,
-		      int    *req_x,
-		      int    *req_y,
-		      int    *req_width,
-		      int    *req_height)
+dialog_constrain_geometry(Client *c,
+			  int    *req_x,
+			  int    *req_y,
+			  int    *req_width,
+			  int    *req_height)
 {
   Wm  *w = c->wm;
   int avail_x, avail_y, avail_width, avail_height;
@@ -454,11 +449,12 @@ dialog_check_geometry(Client *c,
   if (w->config->dialog_stratergy == WM_DIALOGS_STRATERGY_FREE)
     return True;
 
-  /* Allow decorationless dialogs to position themselves anywhere */
-  if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
+  /* Allow urgent dialogs to position themselves anywhere */
+  if (c->flags & CLIENT_HAS_URGENCY_FLAG)
     return True;
 
-  if (c->flags & CLIENT_IS_MESSAGE_DIALOG)
+  /* Decorationless dialogs position anywhere */
+  if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
     return True;
 
   dialog_get_available_area(c,&avail_x, &avail_y, &avail_width, &avail_height);
@@ -539,9 +535,21 @@ dialog_init_geometry(Client *c)
   if (w->config->dialog_stratergy == WM_DIALOGS_STRATERGY_FREE)
     return;
 
-  /* Allow decorationless dialogs to position themselves anywhere */
+  /* Allow decorationless dialogs to position themselves anywhere 
+   * But centered initially if 0,0 
+  */
   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
-    return;
+    {
+      if (c->x == 0 && c->y == 0)
+	{
+	  if (c->height < w->dpy_height)
+	    c->y = (w->dpy_height - c->height)/2;
+
+	  if (c->width < w->dpy_width)
+	    c->x = (w->dpy_width - c->width)/2;
+	}
+      return;
+    }
 
   dialog_get_available_area(c,&avail_x, &avail_y, &avail_width, &avail_height);
 
@@ -559,7 +567,7 @@ dialog_init_geometry(Client *c)
   /* Message Dialogs are free to postion/size where ever but can use totally  
    * offscreen request to position to window corners - see below
    */
-  if (c->flags & CLIENT_IS_MESSAGE_DIALOG)
+  if (c->flags & CLIENT_HAS_URGENCY_FLAG)
     {
       int win_width  = c->width + bdr_east;
       int win_height = c->height + bdr_south;
@@ -618,12 +626,16 @@ dialog_init_geometry(Client *c)
 void
 dialog_client_configure(Client *c)
 {
+  dbg("%s() client has border only hint: %s\n",
+      __func__, (c->flags & CLIENT_BORDERS_ONLY_FLAG) ? "yes" : "no");
   dialog_init_geometry(c);
 }
 
 void
 dialog_client_redraw(Client *c, Bool use_cache)
 {
+  Wm *w = c->wm;
+
   Bool is_shaped = False;
 
   int offset_north = 0, offset_south = 0, offset_west = 0, offset_east = 0;
@@ -636,7 +648,7 @@ dialog_client_redraw(Client *c, Bool use_cache)
 
   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG) return;
 
-  if (use_cache && c->backing != None) return;
+  if (use_cache) return;
 
   offset_north = dialog_client_title_height(c);
 
@@ -651,7 +663,8 @@ dialog_client_redraw(Client *c, Bool use_cache)
     frame_ref_top   = FRAME_DIALOG_NORTH;
 
   /* 'message dialogs have there own decorations */
-  if (c->flags & CLIENT_IS_MESSAGE_DIALOG)
+  if (c->flags & CLIENT_HAS_URGENCY_FLAG
+      && theme_has_message_decor(w->mbtheme))
     {
       frame_ref_top   = FRAME_MSG;
       frame_ref_east  = FRAME_MSG_EAST;
@@ -664,40 +677,26 @@ dialog_client_redraw(Client *c, Bool use_cache)
 
 
   is_shaped = theme_frame_wants_shaped_window( c->wm->mbtheme, frame_ref_top);
-  
-  if (c->backing == (Pixmap)NULL)
-    client_init_backing(c, total_w, total_h);
 
   if (is_shaped) client_init_backing_mask(c, total_w, c->height, 
 					  offset_north, offset_south,
 					  offset_east, offset_west);
-#ifdef STANDALONE
-  /* Should prevent some flicker */
-  XSetForeground(c->wm->dpy, c->wm->mbtheme->gc, 
-		 BlackPixel(c->wm->dpy, c->wm->screen));
-  XFillRectangle(c->wm->dpy, c->backing, c->wm->mbtheme->gc, 
-		 0, 0, total_w, total_h);
-#endif
 
-  theme_frame_paint(c->wm->mbtheme, c, frame_ref_top, 
-		    0, 0, total_w, offset_north); 
+  theme_frame_paint(c->wm->mbtheme, c, frame_ref_top, total_w, offset_north); 
     
-  theme_frame_paint(c->wm->mbtheme, c, frame_ref_west, 
-		    0, offset_north, offset_west, c->height); 
+  theme_frame_paint(c->wm->mbtheme, c, frame_ref_west,
+		    offset_west, c->height); 
   
   theme_frame_paint(c->wm->mbtheme, c, frame_ref_east, 
-		    total_w - offset_east, offset_north, 
 		    offset_east, c->height); 
 
   theme_frame_paint(c->wm->mbtheme, c, frame_ref_south, 
-		    0, total_h - offset_south, 
 		    total_w, offset_south); 
 
   /* We dont paint buttons for borderonly and message dialogs */
   if (!(c->flags & CLIENT_BORDERS_ONLY_FLAG
-	|| c->flags & CLIENT_IS_MESSAGE_DIALOG))
+	|| c->flags & CLIENT_HAS_URGENCY_FLAG))
     {
-
       theme_frame_button_paint(c->wm->mbtheme, c, 
 			       BUTTON_ACTION_CLOSE, 
 			       INACTIVE, FRAME_DIALOG, 
@@ -713,7 +712,7 @@ dialog_client_redraw(Client *c, Bool use_cache)
 
     }
 
-  /* XXXX ifdef HAVE_SHAPE */
+
   if (is_shaped && !c->is_argb32)
     {
       XRectangle rects[1];
@@ -723,77 +722,143 @@ dialog_client_redraw(Client *c, Bool use_cache)
       rects[0].width = total_w - offset_west - offset_east;
       rects[0].height = total_h - offset_south - offset_north;
 
-      XShapeCombineRectangles ( c->wm->dpy, c->title_frame, 
+
+      XShapeCombineRectangles ( c->wm->dpy, c->frame, 
 				ShapeBounding,
 				0, 0, rects, 1, ShapeSet, 0 );
-      
+
 #ifndef USE_COMPOSITE
       if (c->wm->config->dialog_shade && (c->flags & CLIENT_IS_MODAL_FLAG)) 
 	{
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
-			     ShapeBounding, 0, 0, 
-			     c->backing_masks[MSK_NORTH], ShapeUnion);
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
-			     ShapeBounding, 0, total_h - offset_south, 
-			     c->backing_masks[MSK_SOUTH], ShapeUnion);
+	  /* client->frame is our lowlighted window, so we only shape
+           * our decor frames. We dont need to do this for composite. 
+	   *
+           * TODO: the logic for all this is very messy. fix.
+	  */
 
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
-			     ShapeBounding, 0, offset_north, 
-			     c->backing_masks[MSK_WEST], ShapeUnion);
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[NORTH], 
+			     ShapeBounding, 0, 0, 
+			     c->backing_masks[MSK_NORTH], ShapeSet);
+
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[SOUTH], 
+			     ShapeBounding, 0, 0,
+			     c->backing_masks[MSK_SOUTH], ShapeSet);
+
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[WEST], 
+			     ShapeBounding, 0, 0,
+			     c->backing_masks[MSK_WEST], ShapeSet);
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[EAST], 
 			     ShapeBounding, 
-			     total_w - offset_east, offset_north,
-			     c->backing_masks[MSK_EAST], ShapeUnion);
+			     0, 0,
+			     c->backing_masks[MSK_EAST], ShapeSet);
 	}
       else
 #endif
 	{
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
+
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[NORTH], 
 			     ShapeBounding, 0, 0, 
-			     c->backing_masks[MSK_NORTH], ShapeUnion);
+			     c->backing_masks[MSK_NORTH], ShapeSet);
 
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
-			     ShapeBounding, 0, total_h - offset_south, 
-			     c->backing_masks[MSK_SOUTH], ShapeUnion);
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[SOUTH], 
+			     ShapeBounding, 0, 0,
+			     c->backing_masks[MSK_SOUTH], ShapeSet);
 
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
-			     ShapeBounding, 0, offset_north, 
-			     c->backing_masks[MSK_WEST], ShapeUnion);
-	  XShapeCombineMask( c->wm->dpy, c->title_frame, 
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[WEST], 
+			     ShapeBounding, 0, 0,
+			     c->backing_masks[MSK_WEST], ShapeSet);
+	  XShapeCombineMask( c->wm->dpy, c->frames_decor[EAST], 
 			     ShapeBounding, 
-			     total_w - offset_east, offset_north,
-			     c->backing_masks[MSK_EAST], ShapeUnion);
+			     0, 0,
+			     c->backing_masks[MSK_EAST], ShapeSet);
 
 	  XShapeCombineShape ( c->wm->dpy, 
 			       c->frame,
 			       ShapeBounding, 0, 0, 
-			       c->title_frame,
-			       ShapeBounding, ShapeSet);
+			       c->frames_decor[NORTH],
+			       ShapeBounding, ShapeUnion);
+	  
+	  XShapeCombineShape ( c->wm->dpy, 
+			       c->frame,
+			       ShapeBounding, 0, total_h - offset_south, 
+			       c->frames_decor[SOUTH],
+			       ShapeBounding, ShapeUnion);
+	  
+	  XShapeCombineShape ( c->wm->dpy, 
+			       c->frame,
+			       ShapeBounding, 0, offset_north,
+			       c->frames_decor[WEST],
+			       ShapeBounding, ShapeUnion);
+	  
+	  XShapeCombineShape ( c->wm->dpy, 
+			       c->frame,
+			       ShapeBounding, 
+			       total_w - offset_east, offset_north,
+			       c->frames_decor[EAST],
+			       ShapeBounding, ShapeUnion);
+
 	}
     }
-
-#ifdef STANDALONE
-   XSetWindowBackgroundPixmap(c->wm->dpy, c->title_frame, c->backing);
-#else
-   XSetWindowBackgroundPixmap(c->wm->dpy, c->title_frame, 
-			      mb_drawable_pixmap(c->backing));
-#endif
-
-   XClearWindow(c->wm->dpy, c->title_frame);
-
-   XFlush(c->wm->dpy);
 }
 
 void
 dialog_client_button_press(Client *c, XButtonEvent *e)
 {
+  Wm *w = c->wm;
   int offset_north = dialog_client_title_height(c);
   int offset_south = 0, offset_west = 0, offset_east = 0;
 
-  if (c->flags & CLIENT_IS_MESSAGE_DIALOG) return;
-
   dialog_client_get_offsets(c, &offset_east, &offset_south, &offset_west);
 
+  if (w->config->dialog_stratergy == WM_DIALOGS_STRATERGY_STATIC)
+    {
+      /* For static undraggble/stack fixed dialog we simply  
+       * hide the dialog when titlebar is clicked on. 
+       *
+       * TODO: Ideally this code should go in drag loop. but
+       *       it seems the servergrab() stops underlying app
+       *       from repainting itself. Need to look more into 
+       *       this.
+       *       Is it safe to simply remove the server grab ?
+       *
+       */
+      if (c->flags & CLIENT_HAS_URGENCY_FLAG)
+	return; 		/* No Effect for message wins */
+
+      if (XGrabPointer(w->dpy, w->root, False,
+		       ButtonPressMask|ButtonReleaseMask,
+		       GrabModeAsync,
+		       GrabModeAsync, 
+		       None, w->curs, CurrentTime) == GrabSuccess)
+	{
+	  XUnmapWindow(w->dpy, c->frame);
+	  _draw_outline(c, c->x - offset_west, c->y - offset_north,
+			c->width + offset_west + offset_east,
+			c->height + offset_north + offset_south);
+	  
+	  XFlush(w->dpy);
+	  
+	  for (;;) 
+	    {
+	      XEvent xev;
+	      XMaskEvent(w->dpy,ButtonReleaseMask|ButtonPressMask, &xev);
+	      if (xev.type == ButtonRelease)
+		{
+		  _draw_outline(c, c->x - offset_west, c->y - offset_north,
+				c->width + offset_west + offset_east,
+				c->height + offset_north + offset_south);
+		  
+		  XMapWindow(w->dpy, c->frame);
+		  break;
+		}
+	    }
+	  
+	  XUngrabPointer(w->dpy, CurrentTime); 
+	  XSync(w->dpy, False);
+	}
+      return;
+    }
+
   switch (client_button_do_ops(c, e, FRAME_DIALOG, 
 			       c->width + offset_east + offset_west, 
 			       offset_north))
@@ -809,7 +874,9 @@ dialog_client_button_press(Client *c, XButtonEvent *e)
       case -1: 		 /* Cancelled  */
 	 break;
       case 0:
-	 dialog_client_drag(c);     /* Not on button */
+	if (w->config->dialog_stratergy == WM_DIALOGS_STRATERGY_STATIC)
+	  break; 		/* no dragging, raising */
+	dialog_client_drag(c);  /* Not on button */
 	 break;
    }
 }
@@ -818,17 +885,10 @@ static void
 dialog_client_drag(Client *c) /* drag box */
 {
   XEvent ev;
-  int x1, y1;
-  int old_cx = c->x;
-  int old_cy = c->y;
-    
-  int frm_size     = dialog_client_title_height(c);
-  int offset_south = 0, offset_west = 0, offset_east = 0;
-  int have_grab = 0;
-
-#ifdef USE_MSG_WIN
-  Client *t = NULL;
-#endif 
+  int    x1, y1, old_cx = c->x, old_cy = c->y;
+  int    frm_size     = dialog_client_title_height(c);
+  int    offset_south = 0, offset_west = 0, offset_east = 0;
+  int    have_grab = 0;
 
   dbg("%s called\n", __func__);
 
@@ -841,51 +901,14 @@ dialog_client_drag(Client *c) /* drag box */
       != GrabSuccess)
     return;
 
-  /* Let the comp know theres gonna be damage in out old position.
-   * XXX Must be a better way need to figure it out.    
-   */
-
-
-
-#if (DIALOG_WANT_HIDDEN_DRAG) 	/* hide the dialog on drag */
-
-#ifdef USE_COMPOSITE
-  comp_engine_client_hide(c->wm, c);
-  comp_engine_render(c->wm, c->wm->all_damage);
-#endif
-
-#else
-
  comp_engine_client_show(c->wm, c); 
 
-#endif /* DIALOG_WANT_HIDDEN_DRAG */
-
   c->flags |= CLIENT_IS_MOVING;
 
   _get_mouse_position(c->wm, &x1, &y1);
 
- 
-#if (DIALOG_WANT_HIDDEN_DRAG) 	/* hide the dialog on drag */
-
-#ifndef USE_COMPOSITE 		/* .. for lowlighted dialogs */
-  if (c->flags & CLIENT_IS_MODAL_FLAG && c->wm->config->dialog_shade)
-    {
-      XUnmapWindow(c->wm->dpy, c->window);
-      XUnmapWindow(c->wm->dpy, c->title_frame);
-    }
-  else 
-#endif
-    {
-      XUnmapWindow(c->wm->dpy, c->frame);
-    }
-
-  c->ignore_unmap++;
-#endif  /* DIALOG_WANT_HIDDEN_DRAG */
-
   XFlush(c->wm->dpy);
 
-
-  
   for (;;) 
     {
       int wanted_x = 0, wanted_y = 0;
@@ -914,28 +937,23 @@ dialog_client_drag(Client *c) /* drag box */
 		      c->width + offset_west + offset_east,
 		      c->height + frm_size + offset_south);
 
-#if (DIALOG_DRAG_MODE)
-
 	wanted_x = (old_cx + (ev.xmotion.x - x1));
 	wanted_y = (old_cy + (ev.xmotion.y - y1));
 
 	switch (DIALOG_DRAG_MODE) 
 	  {
-	  case DIALOG_DRAG_NONE:
-	    /* No modifications to postion */
-	    break;
 	  case DIALOG_DRAG_RESTRAIN:
 
-	    if ( (wanted_x - offset_west) < 0 - DIALOG_DRAG_RESTRAIN_BDR)
-	      c->x = 0 - DIALOG_DRAG_RESTRAIN_BDR + offset_west;
-	    else if ( (wanted_x + c->width + offset_east) > c->wm->dpy_width + DIALOG_DRAG_RESTRAIN_BDR)
-	      c->x = c->wm->dpy_width + DIALOG_DRAG_RESTRAIN_BDR - (c->width + offset_east);
+	    if ( (wanted_x - offset_west) < 0)
+	      c->x = offset_west;
+	    else if ( (wanted_x + c->width + offset_east) > c->wm->dpy_width)
+	      c->x = c->wm->dpy_width - (c->width + offset_east);
 	    else c->x = wanted_x;
 	    
-	    if ( (wanted_y - frm_size) < 0 - DIALOG_DRAG_RESTRAIN_BDR)
-	      c->y = 0 - DIALOG_DRAG_RESTRAIN_BDR + frm_size;
-	    else if ( (wanted_y + c->height + offset_south) > c->wm->dpy_height + DIALOG_DRAG_RESTRAIN_BDR)
-	      c->y = c->wm->dpy_height + DIALOG_DRAG_RESTRAIN_BDR - (c->height + offset_south);
+	    if ( (wanted_y - frm_size) < 0)
+	      c->y = frm_size;
+	    else if ( (wanted_y + c->height + offset_south) > c->wm->dpy_height)
+	      c->y = c->wm->dpy_height - (c->height + offset_south);
 	    else c->y = wanted_y;
 
 	    break;
@@ -947,15 +965,6 @@ dialog_client_drag(Client *c) /* drag box */
 	    break;
 	  }
 
-
-#else  /* Dialog drag mode disabled below */
-
-	if (c->wm->config->dialog_stratergy != WM_DIALOGS_STRATERGY_CONSTRAINED_HORIZ)
-	  c->x = old_cx + (ev.xmotion.x - x1);
-
-	c->y = old_cy + (ev.xmotion.y - y1);
-#endif  /* DIALOG_WANT_HIDDEN_DRAG */
-
 	_draw_outline(c, c->x - offset_west, c->y - frm_size,
 		      c->width + offset_west + offset_east,
 		      c->height + frm_size + offset_south);
@@ -970,12 +979,6 @@ dialog_client_drag(Client *c) /* drag box */
 #ifndef USE_COMPOSITE	
 	if (c->wm->config->dialog_shade && (c->flags & CLIENT_IS_MODAL_FLAG))
 	  {
-	    XMoveResizeWindow(c->wm->dpy, 
-			      c->title_frame, 
-			      c->x - offset_west, 
-			      c->y - frm_size, 
-			      c->width + offset_east + offset_west,
-			      c->height + frm_size + offset_south);
 	    XMoveResizeWindow(c->wm->dpy, 
 			      c->window, 
 			      c->x, 
@@ -992,31 +995,19 @@ dialog_client_drag(Client *c) /* drag box */
 	
 	c->show(c);
 	
-#ifdef USE_MSG_WIN
-	/* Check for message super above dialogs */
-	START_CLIENT_LOOP(c->wm, t)
-	  {
-	    if (t->flags & CLIENT_IS_MESSAGE_DIALOG
-		&& !(t->flags & CLIENT_IS_MESSAGE_DIALOG_HI)
-		&& !(t->flags & CLIENT_IS_MESSAGE_DIALOG_LO))
-	      {
-		t->show(t);
-		comp_engine_client_show(c->wm, t); 
-		break;
-	      }
-	  }
-	END_CLIENT_LOOP(c->wm, t);
-#endif /* USE_MSG_WIN */
 	c->flags &= ~ CLIENT_IS_MOVING;
 	
 	XUngrabPointer(c->wm->dpy, CurrentTime);
 	XUngrabServer(c->wm->dpy);
+	wm_activate_client(c);
 	return;
       }
     }
 
   XUngrabPointer(c->wm->dpy, CurrentTime);
   XUngrabServer(c->wm->dpy);
+
+
   client_deliver_config(c);
 }
 
@@ -1041,43 +1032,76 @@ _draw_outline(Client *c, int x, int y, int width, int height)
 		 x-1, y-1, width+2, height+2);
 
 }
- 
-void dialog_client_destroy(Client *c)
+
+static Client*
+dialog_client_set_focus_next(Client *c)
+{
+  Wm *w = c->wm; 
+
+  if (c->next_focused_client && c != c->next_focused_client)
+    client_set_focus(c->next_focused_client); 
+  else
+    {
+      if (w->focused_client == c)
+	w->focused_client = NULL;
+      return wm_get_visible_main_client(w);
+    }
+
+  return NULL;
+}
+
+void
+dialog_client_iconize(Client *c)
 {
   Wm     *w = c->wm; 
   Client *d = NULL;
-#ifdef USE_MSG_WIN
-  int was_msg = (c->flags & CLIENT_IS_MESSAGE_DIALOG
-		 && !(c->flags & CLIENT_IS_MESSAGE_DIALOG_HI)
-		 && !(c->flags & CLIENT_IS_MESSAGE_DIALOG_LO));
-#endif 
+
+  client_set_state(c, IconicState);
+  c->flags |= CLIENT_IS_MINIMIZED;
+  c->mapped = False;
+  XUnmapWindow(w->dpy, c->frame); 
+  
+  if ((d = dialog_client_set_focus_next(d)) != NULL)
+    wm_activate_client(d);
+}
+ 
+void 
+dialog_client_destroy(Client *c)
+{
+  Wm     *w = c->wm; 
+  Client *d = NULL;
+
+#ifdef MSG_Q
+  int was_msg = 0;
+
+  if (c->flags & CLIENT_IS_MESSAGE_DIALOG)
+    {
+      if (w->msg_win_queue_head 
+	  && w->msg_win_queue_head->win == c->window);
+      was_msg = 1;
+    }
+#endif
+
+
+  /* Focus the saved next or return a likely candidate if none found */
+  d = dialog_client_set_focus_next(c);
 
   base_client_destroy(c);
 
-  /* We could be a dialog over a desktop, in which case we
-     need to give it back keyboard focus - this is a bit
-     ugly. XXX Improve XXX
-  */
+#ifdef MSG_Q
+if (was_msg)
+  {
+    dbg("%s() was message poping queue\n", __func__);
+    wm_msg_win_queue_pop(w);
+  }
+#endif
 
-  if ((d = wm_get_visible_main_client(w)) != NULL)
-    {
-      if (client_want_focus(d))
-	{
-	  XSetInputFocus(w->dpy, d->window,
-			 RevertToPointerRoot, CurrentTime);
-	  w->focused_client = d;
-	}
+  /* 
+   *  We call activate_client mainly to figure out what to focus next.
+   *  This probably only happens in the case of transient for root
+   *  dialogs which likely have no real focus history.
+   */
+  if (d) 
+      wm_activate_client(d);
 
-      if (w->flags & DESKTOP_RAISED_FLAG)
-	comp_engine_client_show (w, d); 
-    }
-
-#ifdef USE_MSG_WIN
-
-   if (was_msg) 
-     {
-       dbg("%s() was message poping queue\n", __func__);
-       wm_msg_win_queue_pop(w);
-     }
-#endif 
 }
